#! /usr/bin/python2
# -*- coding: utf-8 -*-

import sys
from collections import defaultdict

from formed.meta.io.document import openDocument
from formed.meta.data import ChoiceNode, RadioNode, DateLeaf, TextLeaf, \
     RepeatNode

# Note: we import formed.export.sql even though we don't use it directly
# because otherwise the import of formed.export.rules fails due to
# circular imports (at least with libformed 1047:82ed31e1bf78)
import formed.export.sql
from formed.export.rules import recursive_parse


class UnsupportedError(Exception):

    pass


def rg_ancestors(nc):
    return [item for item in reversed(nc.pathToRoot())
            if isinstance(item, RepeatNode)]

SQL = """\
-- This file was automatically generated by scripts/abortcmsql.py.
-- Do not modify this file manually!

CREATE OR REPLACE FUNCTION abort_case(mid int)
RETURNS void AS
$$
DECLARE
    current_phase INTEGER;
BEGIN
    SELECT INTO current_phase m.phase FROM master_tbl_view m WHERE m.id = mid;

    CASE current_phase
        WHEN 2 THEN
	    -- running CM
%(abortcm)s
%(finalisecm)s
	WHEN 4 THEN
	    -- running BER
%(abortber)s
%(finaliseber)s
	ELSE
    END CASE;
END;
$$ LANGUAGE plpgsql VOLATILE SECURITY DEFINER;

REVOKE ALL ON FUNCTION abort_case(int) FROM public;
GRANT EXECUTE ON FUNCTION abort_case(int) TO GROUP :admin_group;
"""

class Field(object):

    def __init__(self, name, value, rules=(), conditions=()):
        self.name = name
        self.value = value
        self.rules = rules
        self.conditions = conditions
        self.dependencies = set(name for cond in self.rules + self.conditions
                                for name in cond.getExpr().depends)

    def __repr__(self):
        return ("Field(%r, %r, %r, %r))"
                % (self.name, self.value, self.rules, self.conditions))


special_fields_cm = [Field("datum_cm_ende", "CURRENT_DATE"),
                     Field("art_cm_ende", "0"),
                     Field("abbruch_durch", "5"),
                     Field("gruende_abbruch_cm", "9"),
                     Field("phase", "3"),
                     ]

special_fields_ber = [Field("datum_ber_ende", "CURRENT_DATE"),
                      Field("phase", "5"),
                      ]

known_special_fields = set(field.name
                           for field in special_fields_cm + special_fields_ber)


def phase_requirements(formed):
    """Determine required fields for each phase.

    The return value is a dict mapping phases to dicts which in turn map
    table names to lists of Field instances. The table names omit the
    "_tbl" or "_tbl_view" suffix used in the database, so that for
    repeat groups, it's exactly the repeat group name and for the fields
    in master_tbl it's just 'master'. The return value might look like
    this (omitting the details of the lists fields):

       {0: {'master': [...]},
        1: {'master': [...]},
        2: {},
        3: {'master': [...],
            u'rg_biographie2': [...],
            u'rg_ziele': [...]},
        4: {},
        5: {}}

    Limitations:

    Nested repeat groups are not supported. If nested repeat groups
    containing required fields are found, this function raises
    UnsupportedError.
    """
    phase_nc = formed.name2nc["phase"]
    requirements = dict((phase, defaultdict(list))
                        for phase in phase_nc.getValueMapping().keys())

    for name, nc in formed.name2nc.items():
        if name in known_special_fields:
            continue
        rgs = rg_ancestors(nc)
        if not rgs:
            tbl = "master"
        elif len(rgs) == 1:
            tbl = rgs[0].getName()
        else:
            raise UnsupportedError("Field %r has multiple repeat group"
                                   " ancestors" % (nc.getName(),))

        for required_phase in nc.get_parsed_flags().get("required", []):
            phase = int(required_phase[-1])
            requirements[phase][tbl].append(field_from_nc(nc, formed))

    return dict((phase, dict(mapping))
                for (phase, mapping) in requirements.items())


def field_from_nc(nc, formed):
    if isinstance(nc, (ChoiceNode, RadioNode)):
        value = "-3"
    elif isinstance(nc, TextLeaf):
        value = "'unbekannt'"
    elif isinstance(nc, DateLeaf):
        value = "'0001-01-01'"
    else:
        raise UnsupportedError("Cannot determine the 'unknown' value for %r"
                               % (nc,))

    return Field(nc.getName(), value, *get_conditions(nc, formed))


def get_rules(nc, formed):
    return [rule for rule in formed.name2markingrules.get(nc.getName(), ())
            if rule.isWarning() and rule.getName().startswith("required")]


def get_conditions(nc, formed):
    rules = []
    conditions = []
    for rule in get_rules(nc, formed):
        rules.append(rule)
        for var in rule.getExpr().getDependencies():
            for conditional in formed.name2conditionals[var]:
                conditions.append(conditional)
    return (rules, conditions)


def format_condition(field, formed):
    conditionals = " AND ".join(recursive_parse(cond.getExpr().prog, formed)
                                for cond in field.conditions)
    rules = " AND ".join(recursive_parse(cond.getExpr().prog, formed)
                         for cond in field.rules)
    if rules:
        rules = "NOT (" + rules + ")"
    return " AND ".join(sql for sql in (conditionals, rules) if sql)


def sort_fields(fields):
    """Sort fields based on their dependencies.

    Some of the required fields have conditions that depend on other
    required fields. To make sure the data is consistent after all the
    updates we want to make sure to update the fields other fields'
    conditions depend on before those other fields. This function sorts
    the fields by that criterion (it effectively performs a topological
    search).
    """
    if not fields:
        return []

    fieldmap = dict((field.name, field) for field in fields)

    processed = set()
    result = []
    todo = fields
    while todo:
        field = todo.pop(0)
        if field.name in processed:
            continue

        deps = [fieldmap[dep] for dep in field.dependencies
                if dep != field.name
                if dep in fieldmap
                if dep not in processed]
        if deps:
            todo.append(field)
            todo.extend(deps)
        else:
            result.append(field)
            processed.add(field.name)
    return result


def make_update_statements(tablefields, formed):
    statements = []
    for table, fields in sorted(tablefields.items()):
        fields = sort_fields(fields)
        if not fields:
            return ""

        setters = []
        for field in fields:
            condition = format_condition(field, formed)
            statements.append("UPDATE %s_tbl_view SET %s = %s"
                              "\n WHERE %s = mid%s%s;"
                              % (table, field.name, field.value,
                                 "id" if table == "master" else "master_id",
                                 "\n   AND " if condition else "",
                                 condition))
    return "\n".join(12 * " " + line
                     for statement in statements
                     for line in statement.split("\n"))


def format_sql(requirements, formed):
    return SQL % dict(abortcm=make_update_statements(requirements[3], formed),
                      finalisecm=make_update_statements(
                          dict(master=special_fields_cm), formed),
                      abortber=make_update_statements(requirements[5], formed),
                      finaliseber=make_update_statements(
                          dict(master=special_fields_ber), formed))


def main():
    formed = openDocument(sys.argv[1], generate_rules=True)
    requirements = phase_requirements(formed)
    print format_sql(requirements, formed)

main()
