# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008, 2010, 2014 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD.
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls JMD has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
#

import logging

from formed.instance.tree import SetDataException

from mpulsweb.lib.base import g
from mpulsweb.model.case import MpulsCase, MpulsCaseFactory, \
     MpulsCaseOverview, MpulsCaseBundle, MpulsCaseState


log = logging.getLogger(__name__)


class CaseBundle(MpulsCaseBundle):
    pass


class CaseOverview(MpulsCaseOverview):
    pass


class State(MpulsCaseState):
    pass


class CaseFactory(MpulsCaseFactory):
    pass


class Case(MpulsCase):

    def create_repeat_group_instance(self, rg_id):
        """Overwrite inherited method to set kopie_geprueft checkboxes.
        Some repeat groups have leaves for checkboxes that need to be
        checked by the user to indicate that the contents have been
        verified when a new structure case has been created by copying
        an old one. For newly created repeat group instances, this isn't
        really needed because their contents are not the result of
        copying. However, libformed cannot really set a useful default
        value for this field and the one it actually sets does not work
        well with the required field rules (the default value is -1 and
        that is considered to be a value set by the user because it's
        not NULL/None).

        As a work-around this method sets the value of these leaves in
        all new repeat group instances to 0, which normally means 'No'
        but we interpret it as the RG instance not being the result of
        copying. The conditional in the form hides the checkbox in this
        case.
        """
        instance = super(Case, self).create_repeat_group_instance(rg_id)
        rg_name = instance.meta.getName()
        for leaf_name in g.structureTree.root.repeat_groups[rg_name].leaves:
            if leaf_name.startswith("kopie_geprueft_p"):
                instance_tree = self.get_formed_instance_tree()
                node_name = instance.get_node(leaf_name).getIdentifier()
                try:
                    # We cannot set the value 0 when giving the value as
                    # a string because the setData method maps all empty
                    # strings to None/NULL.
                    instance_tree.setData({node_name: 0}, convert=False)
                except SetDataException, e:
                    # SetDataException should not happen for the values
                    # set here, i.e. for the kopie_geprueft fields, but
                    # they may happen for other fields due to the
                    # "uncommitted fields" mechanism. Therefore it's OK
                    # in this method if we get an Exception for one of
                    # the other fields, but if we get an Exception for
                    # the kopie_geprueft field we're trying to change
                    # here, that's a programming error and we reraise
                    # the exception
                    if node_name in e.errors:
                        raise

                instance_tree.commit()
        return instance


def used_years(cur):
    """Return a list with the years for which statistics exist.
    Statistics that have been marked for deletion are considered to
    exist. Otherwise, if a marking for deletion and restoring could lead
    to the database containing two statistics for the same year.
    """
    cur.execute("SELECT DISTINCT"
                " extract(year from datum_begin_zeitraum) as year"
                " FROM master_tbl_view")
    return [t[0] for t in cur.fetchall()]

def find_unused_years(cur, minyear, maxyear):
    """Return a list of all unused years in range(minyear, maxyear)."""
    used = set(used_years(cur))
    return [year for year in range(minyear, maxyear) if year not in used]


def year_used(cur, year):
    """Determine whether the given year is covered by a statistic yet.
    Uses the same logic to determine whether a year is used as
    used_years.
    """
    cur.execute("SELECT EXISTS"
                " (SELECT * FROM master_tbl_view"
                "  WHERE extract(year from datum_begin_zeitraum) = %s)",
                (year,))
    return cur.fetchone()[0]


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
