# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls JMD has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
# Sascha L. Teichmann <teichmann@intevation.de>
#

import logging
import datetime
import json

from pylons import tmpl_context as c, app_globals as g, request, url
from pylons.controllers.util import redirect, abort

import mpulsweb.lib.helpers as h
from mpulsweb.lib.dialogs import checkPrivacyStatement
from mpulsweb.lib.security import checkRole, hasRole, check_state
from mpulsweb.lib.base import session, render, restrict_to_POST, \
    abort_unless_POST
from mpulsweb.lib.translation import _
from mpulsweb.controllers.case import CaseController as MpulsCaseController
from mpulsweb.model.agencysettings import Agency
from mpulsweb.model.case import OPEN_STATE

from jmdweb.controllers.formularpage import JMD2StartRendererController
from jmdweb.controllers.case_bundle import build_case_list
from jmdweb.lib.validators import NewCaseForm
from jmdweb.model.case import JMDXMLImporter, search_duplicates


log = logging.getLogger(__name__)


class DuplicatesFound(Exception):

    def __init__(self, form_result, duplicates):
        self.form_result = form_result
        self.duplicates = duplicates


class CaseController(JMD2StartRendererController, MpulsCaseController):

    def _get_validator(self):
        return NewCaseForm()

    def _meta_tree_for_imports(self):
        return g.formed_tree_for_imports

    def _get_xmlimporter(self, xmlfile=None):
        '''This function should be overwritten in specific applications to add
        a custom XML Importer'''
        allow_documents = g.mpuls_config.is_enabled('case-module', 'documents')
        allow_logbook = g.mpuls_config.is_enabled('case-module', 'logbook')
        return JMDXMLImporter(self._meta_tree_for_imports(),
                              allow_logbook=allow_logbook,
                              allow_documents=allow_documents)

    def _add_custom_fields(self):
        fields = {}

        agency = Agency()
        name = agency.getName()
        if name:
            fields['jmd_name'] = name

        return fields

    # Completely overwrite the _render_new method to unset
    # htmlfill force_defaults.
    def _render_new(self, session_case, form_defaults=None, form_errors=None):
        if form_defaults is None:
            form_defaults = session_case.getData()
        if not form_defaults.has_key("datum_ea_start"):
            form_defaults["datum_ea_start"] = h.format_date(
                                                    datetime.date.today())
        if form_errors is None:
            form_errors = {}

        from pylons import tmpl_context as c
        c.case = session_case
        c.plz_pairs = json.dumps([["adresse_plz", "adresse_ort"]])

        from mpulsweb.lib.base import render, render_htmlfill
        form = render('/casemanagement/new.mako')
        return render_htmlfill(form, defaults=form_defaults,
                               force_defaults=False,
                               auto_insert_errors=False, errors=form_errors)


    @checkRole('cm')
    @restrict_to_POST
    def newAction(self):
        """Extend newAction to print also the discretion statement."""

        from pylons import request, tmpl_context as c
        from mpulsweb.model.case import SessionCase
        from mpulsweb.lib.base import session, render
        from jmdweb.model.statement import DiscretionStatement
        from jmdweb.model.statement import PrivacyStatement

        session_case = SessionCase()
        session_case.setData(request.params)

        session['case'] = session_case
        # Use the session to store the folding information
        session['folded.EinwilligungenMuster'] = \
                    request.params.get('folded.EinwilligungenMuster')
        session['folded.SchweigepflichtMuster'] = \
                    request.params.get('folded.SchweigepflichtMuster')
        session.save()

        # Do we want to print the discretion statement?
        if "printDiscretion" in request.params:
            c.print_version = 1
            statement = DiscretionStatement(session_case, Agency())
            c.content = statement.fillout()
            return render('privacy/default_statement.mako')

        elif "print" in request.params:
            c.print_version = 1
            statement = PrivacyStatement(session_case, Agency())
            c.content = statement.fillout()
            return render('privacy/default_statement.mako')
        else:
            try:
                return super(CaseController, self).newAction()
            except DuplicatesFound as df:
                c.duplicates = df.duplicates
                return self._render_new(session_case, df.form_result)

    def _createCase(self, form_result):
        if not form_result["ignore_duplicates"]:
            name = form_result["name"]
            vorname = form_result["vorname"]
            if name and vorname:
                # check only if both name and vorname have been given.
                duplicates = search_duplicates(name, vorname)
                if duplicates:
                    raise DuplicatesFound(form_result, duplicates)
        return super(CaseController, self)._createCase(form_result)


    @checkRole(('admin', 'cm', 'pt_dlr'))
    # NOTE: checkPrivacyStatement calls self._loadCase which updates the
    # session state including the current case id and the navigation
    # tree. The select method implementation below relies on this!
    @checkPrivacyStatement()
    def select(self, id, confirmed=0):
        """Overwrite base method to redirect to first form page."""
        navigation = session['navigation.tree']
        redirect(url(controller="navigation", action="select_branch",
                     id=navigation.first_page.key))


    def _build_case_list(self, cases):
        """Overwrite base class method to use JMD specific formatter"""
        return build_case_list(cases)

    def _organisation_extra(self, case):
        c.show_abort = hasRole("admin")
        if not c.show_abort:
            return

        if case.is_editable_state():
            phasestate = case.get_state_machine().get_current_state(case)
            if phasestate.get_id() in (2, 4):
                c.can_abort = True
                c.abort_message = (_("The case is in phase %s and can be"
                                     " aborted prematurely.")
                                   % (phasestate.get_display_name(),))
            else:
                c.can_abort = False
                c.abort_message = (_("The case is in phase %s and cannot be"
                                     " aborted prematurely.")
                                   % (phasestate.get_display_name(),))
        else:
            c.can_abort = False
            c.abort_message = _("The case is not in an editable state and"
                                " therefore cannot be aborted prematurely.")

    @checkRole('admin')
    def abort_case(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if not check_state(case, {'admin': [OPEN_STATE]}):
            abort(401)

        if confirmed:
            abort_unless_POST()
            case.abort_case()
            case.markAnonymize()
            c.dialog_title = _("Case document aborted and marked for"
                               " anonymization!")
            c.dialog_text = _("The case document has been aborted and given"
                              " to the administrator for anonymization.")
            c.url_ok = h.url(controller="case_overview", action='index')
            session['render_mode'] = 'ro'
            session.save()
            return render('/casemanagement/dialogs/success_abort.mako')
        else:
            c.context = "../main.mako"
            c.dialog_title = _('Abort case?')
            c.dialog_text = _("Do you really want to abort this case?"
                              " %s The case will be documented as having"
                              " been aborted, the missing required fields will"
                              " be set to 'unknown' and the case will given"
                              " to the administrator for anonymization.") \
                % "\n".join(self._build_case_list([case]))

            c.url_yes = h.url(controller='case', action='abort_case',
                              id=id, confirmed=1)
            c.url_no = h.url(controller='case', action='organisation', id=id)
            return render('/casemanagement/dialogs/confirm.mako')


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
