# -*- coding: utf-8 -*-
#
# Copyright 2015 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls_S (CoMPUter-based case fiLeS -
# Deutschlandstipendium).
#
# mpuls_S is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls_S is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls_S. If not, see <http://www.gnu.org/licenses/>.

import logging
import json

from pylons import tmpl_context as c, request

from formed.renderer.html import Item, GermanFormatter
from formed.renderer.helpers import escape, tag
from formed.meta.semantic import ErrorItem
from formed.instance.tree import SetDataException

from mpulsweb.controllers.formularpage import FormularpageController as \
    MpulsFormularpageController, PageRenderer as MpulsPageRenderer
from mpulsweb.lib.formedrenderer import ViewRenderer as MpulsViewRenderer, \
    ViewPrintRenderer as MpulsViewPrintRenderer
from mpulsweb.lib.base import session
from mpulsweb.lib.translation import _
from mpulsweb.model.agencysettings import Agency

from jmdweb.model.case import search_duplicates


log = logging.getLogger(__name__)


class DuplicateItem(ErrorItem):

    """Version of ErrorItem for use with the duplicate detection

    It holds a list of the duplicate cases as a way of communicating
    them to the renderers. It's a new class also so that it can be
    easily detected in the entire set of uncommitted fields.
    """

    def __init__(self, *args, **kw):
        self.duplicates = kw.pop("duplicates", [])
        super(DuplicateItem, self).__init__(*args, **kw)



class JMD2StartMixin(object):

    """Mix-in class to add jmd2start functionality to a ViewRenderer.

    When using this as a mix-in together with
    mpulsweb.lib.formedrenderer's ViewRenderer or ViewPrintRenderer, the
    jmd2start specific fields will not be rendered.

    Since this class overwrites methods from those other classes it must
    come before those classes in the method resolution order, which can
    be achieved by listing this mix-in before one of those other classes
    in the base class list of a derived class.
    """

    def __init__(self, *args, **kw):
        self.show_jmd2start = kw.pop("show_jmd2start", False)
        super(JMD2StartMixin, self).__init__(*args, **kw)

    def render_nc(self, nc):
        # Hide group "group-zielgruppe" for non-jmd2start agencies
        if not self.show_jmd2start and nc.getName() == "group-zielgruppe":
            return Item("")
        return super(JMD2StartMixin, self).render_nc(nc)


class ViewRenderer(JMD2StartMixin, MpulsViewRenderer):

    """A ViewRenderer that includes the JMD2StartMixin"""

    def __init__(self, *args, **kw):
        self.duplicate_items = kw.pop("duplicate_items", {})
        super(ViewRenderer, self).__init__(*args, **kw)

    def render_page(self, *args):
        c.plz_pairs = json.dumps([["adresse_plz", "adresse_ort"],
                                  ["inst_plz", "inst_ort"]])
        return (self.render_duplicates()
                + super(ViewRenderer, self).render_page(*args))


    def render_duplicates(self):
        """Render a box with information about duplicate cases."""
        duplicates = []
        for item in self.duplicate_items:
            duplicates = item.duplicates

        if not duplicates:
            return []

        name = self.data.get_value("name")
        vorname = self.data.get_value("vorname")

        out = [
            tag("div", class_="form-group mb-4 col-8"),
            tag("div", class_="form-row"),
            tag("div", class_="form-group"),
            tag("div", class_="alert alert-warning"),
            tag("h5"),
            escape(u"Mögliche Duplikate gefunden"),
            "</h5>",
            tag("p"),
            escape(u"Es besteht die Möglichkeit, dass "),
            tag("em"),
            escape(u"%s, %s" % (name, vorname)),
            "</em>",
            escape(u" bereits von Ihrem JMD begleitet wird. Datensätze mit"
                   u" ähnlichen Namen liegen vor. Bitte klären Sie zusammen,"
                   u" ob hier eine doppelte Dokumentation oder eine"
                   u" Namensähnlichkeit vorliegt."),
            "</p>",
            tag("table", class_="table table-sm custom-table-striped"),
            tag("thead"),
            tag("tr"),
            tag("th", class_="table_header_h"),
            escape(_(u'Lastname')),
            "</th>",
            tag("th", class_="table_header_h"),
            escape(_(u'Firstname')),
            "</th>",
            tag("th", class_="table_header_h"),
            escape(u"Bearbeiter*in"),
            "</th>",
            "</tr>",
            "</thead>",
            tag("tbody"),
        ]
        for d in duplicates:
            out.extend([
                tag("tr"),
                tag("th"), escape(d.name), "</th>",
                tag("th"), escape(d.vorname), "</th>",
                tag("th"), escape(d.bearbeiter), "</th>",
                "</tr>",
                ])
        out.extend([
            "</tbody>",
            "</table>",
            tag("input", type="checkbox", name="ignore_duplicates",
                id="ignore_duplicates", value="ignore-duplicates"),
            tag("label", for_="ignore_duplicates"),
            escape(u"Akte trotzdem ändern"),
            "</label>",
            "</div>",
            "</div>",
            "</div>",
            "</div>",
        ])
        return out


class ViewPrintRenderer(JMD2StartMixin, MpulsViewPrintRenderer):

    """A ViewPrintRenderer that includes the JMD2StartMixin"""


class PageRenderer(MpulsPageRenderer):

    """PageRenderer that can handle information about duplicate cases"""

    def convert_error_items(self, error_items):
        error_items, duplicate_defaults = self.extract_duplicates(error_items)
        errors, defaults = super(PageRenderer, self).convert_error_items(error_items)
        defaults.update(duplicate_defaults)
        return errors, defaults

    def extract_duplicates(self, errors):
        """Extract information about the duplicate cases from the error items

        The return vaue is a pair with the remaining error items and a
        dictionary with the default values to use when rendering the
        name fields.
        """
        duplicate_items = []
        new_errors = []
        for items in errors:
            new_items = []
            for item in items:
                if isinstance(item, DuplicateItem):
                    duplicate_items.append(item)
                else:
                    new_items.append(item)
            new_errors.append(new_items)
        self.viewrenderer_args["duplicate_items"] = duplicate_items
        duplicate_defaults = dict((item.name, item.bad)
                                  for item in duplicate_items)
        return new_errors, duplicate_defaults



class JMD2StartRendererController(object):

    """Mix-in class for controllers that need to use jmd2start renderers.

    This mix-in overwrites the _get_formpage_renderer method to return a
    page renderer that only renders the jmd2start fields for agencies
    that take part in jmd2start.
    """

    def _is_jmd2start(self):
        return Agency().get_value("jmd2start")

    def _get_formpage_renderer(self):
        show_jmd2start = self._is_jmd2start()
        pagerenderer = PageRenderer()
        pagerenderer.set_viewrenderer(ViewRenderer,
                                      show_jmd2start=show_jmd2start)
        pagerenderer.set_printrenderer(ViewPrintRenderer,
                                       show_jmd2start=show_jmd2start)
        return pagerenderer.get_rendered_page



class FormularpageController(JMD2StartRendererController,
                             MpulsFormularpageController):

    def _before_set_data_hook(self, case, form_result):
        self._before_set_data_handle_aufenthaltspapier(case, form_result)
        self._before_set_data_handle_einreisedatum(case, form_result)
        self._before_set_data_handle_duplicates(case, form_result)

    def _before_set_data_handle_aufenthaltspapier(self, case, form_result):
        instance_tree = case.get_formed_instance_tree()
        # Handle Aufenthaltspapier_bei_Eintritt and Aufenthaltspapier:
        # If Aufenthaltspapier_bei_Eintritt is set for the first time
        # (i.e. is 'keine Angabe' and also Aufenthaltspapier is 'keine Angabe'.
        KEINE_ANGABE = '-1'

        # Check if form_result has "aufenthaltspapier_eintritt", i.e. field
        # is editable and therewith potentially set to reasonable value
        aufenthaltspapier_eintritt = "aufenthaltspapier_eintritt:%s" % case.id
        if form_result.has_key(aufenthaltspapier_eintritt):

            # Get current value "aufenthaltspapier"
            if instance_tree.has_node("aufenthaltspapier"):
                value_aufenthaltspapier = '%d' \
                    % instance_tree.get_node("aufenthaltspapier").value

            # Check if "aufenthaltspapier" is not yet set.
            # Else do not change value
            if value_aufenthaltspapier == KEINE_ANGABE:

                # Get current value "aufenthaltspapier_eintritt"
                if instance_tree.has_node("aufenthaltspapier_eintritt"):
                    value_aufenthaltspapier_eintritt = '%d' \
                        % instance_tree.get_node("aufenthaltspapier_eintritt").value
                    # Check if if "aufenthaltspapier_eintritt" is not yet set.
                    # Only in this case "aufenthaltspapier_eintritt" is newly
                    # set and its new value can be copied to the also unset
                    # "aufenthaltspapier"
                    if value_aufenthaltspapier_eintritt == KEINE_ANGABE:
                        aufenthaltspapier = "aufenthaltspapier:%s" % case.id
                        form_result[aufenthaltspapier] = \
                            form_result[aufenthaltspapier_eintritt]

    def _before_set_data_handle_einreisedatum(self, case, form_result):
        """Fill einreisedatum with birthdate if """
        country_item = case.get_formed_item("geburtsland")
        entrydate_item = case.get_formed_item("einreisedatum")
        birthdate_item = case.get_formed_item("geburtsdatum")

        # Germany is encoded as 29, but make sure it actually is
        DE = 29
        assert country_item.getMeta().getValueMapping()[DE] == "Deutschland"

        # "geburtsland" and "einreisedatum" are on the same form page,

        # We only need to do something if we're on the page containing
        # "einreisedatum", which we can determine by form_result
        # containing the identifier of the entrydate_item. entrydate
        # will not be None in that case and if it's the empty string, we
        # know that the user didn't enter a value yet. The field
        # "geburtsland" is on the same page as "einreisedatum", so we
        # need to get its value from form_result as well. However,
        # "geburtsdatum" is on a different page, so we take it from the
        # case object.
        form_country = form_result.get(country_item.getIdentifier(), "")
        entrydate = form_result.get(entrydate_item.getIdentifier())
        birthdate = birthdate_item.getValue()
        if (form_country == str(DE) and entrydate == ""
            and birthdate is not None):
            form_result[entrydate_item.getIdentifier()] = \
                GermanFormatter().format_date(birthdate_item.getMeta(),
                                              birthdate)

    def _before_set_data_handle_duplicates(self, case, form_result):
        """Check for duplicate cases.

        If the fields 'name' or 'vorname' would be changed, and the
        checkbox 'ignore_duplicates' has not been checked, this method
        determines whether there already are cases with names similar to
        the new name. If any, this method builds new error items for
        them an stores them in the request for later processing when
        rendering the case.
        """
        if form_result.pop("ignore_duplicates", False):
            return

        name_item = case.get_formed_item("name")
        vorname_item = case.get_formed_item("vorname")
        name = form_result.get(name_item.getIdentifier(), "")
        vorname = form_result.get(vorname_item.getIdentifier(), "")

        if not name or not vorname:
            # We don't have both names, so nothing to do.
            return
        if name_item.value != name or vorname_item.value != vorname:
            # At least one of name/vorname has changed, so look for
            # duplicates
            duplicates = search_duplicates(name, vorname, case.id)
            if duplicates:
                uncomitted = {}
                for item, value in [(name_item, name),
                                    (vorname_item, vorname)]:
                    err = DuplicateItem(ErrorItem.RULE, item.getIdentifier(),
                                        u"Mögliches Duplikat", value,
                                        duplicates=duplicates)
                    uncomitted[err.name] = err
                    del form_result[item.getIdentifier()]
                request.jmd_uncommitted_duplicates = uncomitted

    def _anchor_for_save_button(self, default_anchor):
        """Use custom redirection anchors for the 'unterziel' buttons.

        The redirects for the save buttons that for the unterziel-Groups
        can be particularly confusing because all those groups look
        alike (apart from any contents, only the numbers in the labels
        are different). When users press the save button beneath such a
        group after entering some information into that group, this
        would by default lead to the save button itself being at the top
        of the window afterwards and the group below that replacing the
        group the user worked on. This can give the appearance that all
        inputs were lost.

        In order to compensate for this, we redirect to the group above
        the save button instead of the save button itself. The save
        buttons and groups are assumed to have names starting with
        "save-unterziel-" and "group-unterziel-" respectively.
        """

        if default_anchor.startswith("save-unterziel-"):
            return "f_group" + default_anchor[4:]
        return default_anchor

    def _set_data(self, instance_tree, form_result):
        uncommitted_duplicates = getattr(request, "jmd_uncommitted_duplicates", {})
        try:
            instance_tree.setData(form_result)
        except SetDataException:
            pass
        uncommitted = instance_tree.getErrors().copy()
        for key, error in uncommitted_duplicates.items():
            uncommitted.setdefault(key, []).append(error)
            instance_tree.new_values.pop(key, None)
        instance_tree.uncommited = uncommitted
        if uncommitted:
            raise SetDataException(uncommitted)

    def save(self):
        old_case = session.get('case')
        uncomitted = session.get('uncommited_fields')
        if old_case is not None and uncomitted is not None:
            uncomitted.pop("name:%d" % (old_case.id), None)
            uncomitted.pop("vorname:%d" % (old_case.id), None)
            session["uncommited_fields"] = uncomitted
        return super(FormularpageController, self).save()
