# -*- coding: utf-8 -*-

import logging
from pylons.controllers.util import redirect

from mpulsweb.lib.translation import _

from formed.instance.tree import SetDataException

from mpulsweb.lib.security import checkRole
from mpulsweb.lib.base import BaseController, render, c, h, session, \
    abort_unless_POST
from mpulsweb.model.phase import ConsistenceCheckException, \
    PhaseAnonymizableCheckException
from mpulsweb.controllers.case_overview import CaseOverviewController

log = logging.getLogger(__name__)


class PrivacyController(BaseController):

    '''
    Controller to handle privacy related things within mpuls. This mainly
    includes accepting or declining the 'Childs declaration of consent' (known
    in WASKA/WASKO) or similar. Additionally it is possible to list available
    static statements.
    '''
    @checkRole('cm')
    def index(self):
        return self.listStatements()

    @checkRole('cm')
    def listStatements(self, id=None):
        """Render a list of available static statements related to privacy.

        The id parameter is actually mandatory and has a default value
        only to make sure that clients that continue to fetch
        listStatements (some clients fetch it without first opening a
        case at all via a different request, see mpuls/issue7538)
        without it do not generate Iternal Server Errors (from a
        typeError because of the missing argument) but get a 404
        response code instead.
        """
        id = self._checkInt(id)
        c.case = self._loadCase(id, read_only=True)
        return render('privacy/list_statements.mako')

    @checkRole('cm')
    def printStatement(self, id):
        id = self._checkInt(id)
        case = self._loadCase(id, read_only=True)
        statement = case.getPrivacyStatement()
        c.content = statement.fillout()
        return render('privacy/default_statement.mako')

    @checkRole('cm')
    def set(self, id):
        c.targeturl = h.url(controller="case", action="digest", id=id)
        c.ds_id = id
        return render('privacy/edit_statement.mako')

    @checkRole('cm')
    def accept(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if confirmed:
            abort_unless_POST()
            statement = case.getPrivacyStatement()
            statement.sign()
            # TODO: Think about refreshing the state of the privacy
            # statement in the statement model
            session_case = session.get('case')
            session_case.privacy_statement = True
            session['success_text'] = _(u"Privacy statement set!")
            session.save()
            return redirect(h.url(controller="case", action="select", id=case.id))
        else:
            c.dialog_title = _("Set declaration of consent?")
            c.dialog_text = _("You have stated that there is a declaration of "
                              "consent from\nthe young adult to the "
                              "acquisition, handling "
                              "and storage of personal data.\n"
                              "Are you sure this is correct?")
            c.url_yes = h.url(controller="privacy", action="accept",
                              id=id, confirmed="1")
            c.url_no = h.url(controller="case", action="select",
                             id=id, confirmed=0)
            return render('privacy/dialogs/confirm.mako')

    @checkRole('cm')
    def decline(self, id, confirmed=False):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        try:
            case.check_anonymizeability()
        except PhaseAnonymizableCheckException, e:
            c.dialog_title = _('Decline privacy statement not possible')
            c.dialog_text = h.literal(
                _('The case is in a non consistent state und it will not be '
                  'possible to anonimise it after retracting the privacy '
                  'statement. The reason for this is listed below:%s') % e.value
            )
            c.url_ok = h.url(controller='case', action='digest',
                             id=case.id)
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

        except ConsistenceCheckException, e:
            c.dialog_title = _('Decline privacy statement not possible')
            c.dialog_text = h.literal(
                _('<p>The case is in a non consistent state und it will not be '
                  'possible to anonimise it after retracting the privacy '
                  'statement. The reason for this is listed below:'
                  '</p><p>%s</p><p>Please refer to the documentation in case '
                  'that the privacy statement has been declined.</p>') % e.value
            )
            c.url_ok = h.url(controller='case', action='digest',
                             id=case.id)
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

        if confirmed:
            abort_unless_POST()
            # Mark case for anonymizion
            try:
                case.markAnonymize()
                c.success_text = _(u"The case document has been freed to be"
                                  u" anonymised and has been given to"
                                  u" the admininstration.")
                statement = case.getPrivacyStatement()
                statement.decline()
                return CaseOverviewController().index()
            except SetDataException, e:
                errors = []
                for k, values in e.errors.iteritems():
                    for v in values:
                        errors.append(v.msg)
                c.dialog_title = _(u"""Anonymising not possible.""")
                c.dialog_text = _("To anonymise this case document was not "
                                  "possible, because of following \nreason:"
                                  "</p><p><i>%s</i></p><p>Please click on OK "
                                  "to get back to the overview.\n</p>") % \
                    "<br>".join(errors)
                c.url_ok = h.url(controller='case', action='select',
                                 id=case.id, confirmed=1)
                return render('/privacy/dialogs/decline_failed.mako')

        else:
            # Get description of the phase which will be set after anonymisation
            c.dialog_title = _(u"Retract privacy statement?")
            c.dialog_text = _(
                u"The case will be anonymised if you retract the privacy "
                u"statement.")
            c.url_yes = h.url(controller="privacy", action="decline",
                              id=id, confirmed="1")
            c.url_no = h.url(controller="case", action="select",
                             id=id, confirmed=0)
            return render('privacy/dialogs/confirm.mako')

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8:
