# -*- coding: utf-8 -*-

import logging
from decorator import decorator

from webob.exc import HTTPNotFound

from pylons import request, tmpl_context as c, app_globals as g
from mpulsweb.lib.translation import _

from mpulsweb.lib.base import render, session
from mpulsweb.lib.helpers import url
from mpulsweb.model.appointment import CaseAppointmentFactory, \
     MissingAppointmentError


DIALOGS_ENABLED = True

log = logging.getLogger(__name__)

# MISSING STATEMENT DIALOG
# ------------------------
# If the statement of the declaration of consent is missing on openening a case
# we will open a dialog which informs the user about this fact


def render_privacy_statement_dialog_if_unsigned(case):
    statement = case.getPrivacyStatement()
    # Check if there is a "Einverstaendniserklaerung" None if it is
    # not either signed or declined
    if not statement.isSigned() is None:
        return None
    c.ds_id = case.id
    return render('/privacy/dialogs/missing_statement.mako')


def checkPrivacyStatement(context='select'):
    '''Decorator for the methods in the case controller'''
    def validate(func, self, *args, **kwargs):
        if not g.mpuls_config.is_enabled('case-module', 'privacy'):
            log.debug('Privacy disabled... No checks on statement done')
            return func(self, *args, **kwargs)
        # load case and check if the statement is signed
        if context == "showAppointment":
            try:
                app_id = int(args[0])
                confirmed = int(args[1])
            except:
                raise HTTPNotFound()
            app_factory = CaseAppointmentFactory()
            try:
                app = app_factory.loadById(app_id)
            except MissingAppointmentError:
                raise HTTPNotFound()
            case_id = app.case_id
            c.targeturl = url(controller="caseappointment",
                              action="show",
                              id=app_id, confirmed=1)
        else:
            try:
                case_id = int(args[0])
                confirmed = int(args[1])
            except:
                raise HTTPNotFound()
            c.targeturl = url(controller="case", action=context,
                              id=case_id, confirmed=1)

        case = self._loadCase(case_id)

        if not confirmed:
            privacy_dialog = render_privacy_statement_dialog_if_unsigned(case)
            if privacy_dialog is not None:
                return privacy_dialog

        return func(self, *args, **kwargs)
    return decorator(validate)


# FUTURE CONFIRM DIALOG
# ---------------------
# Not used yet. Use this confirm decaorator in future to set up general
# confirm dialog. Not that this means to loose information in the navigation
# path, as the dialogs do not have any information on this


def get_dialog_template(ctx="case"):
    """
    Depending of the called url this function will return the context
    (global/case) in which the dialog will be displayed
    """
    # TODO: Parse url and return context
    if not ctx:
        ctx = "case"
    return ctx  # All dialogs are shown in case context for now


def render_confirm(header=None, text=None, url_ok=None, url_abort=None,
                   dyntext=None, ctx=None):
    """Returns the rendered confirm dialog"""
    c.dialog_title = header or _('Default Header')
    if dyntext and text:
        c.dialog_text = text % dyntext
    else:
        c.dialog_text = text or _('Default Text')
    c.url_ok = url_ok
    c.url_abort = url_abort or url(controller="meta", action="index")
    return render('/dialogs/confirm_%s.mako' % get_dialog_template(ctx))


def render_success(header=None, text=None, url_ok=None, dyntext=None, ctx=None):
    """Returns the rendered success dialog"""
    c.dialog_title = header or _('Default Header')
    if dyntext and text:
        c.dialog_text = text % dyntext
    else:
        c.dialog_text = text or _('Default Text')
    c.url_ok = url_ok
    return render('/dialogs/success_%s.mako' % get_dialog_template(ctx))


def render_error(header=None, text=None, url_ok=None, dyntext=None, ctx=None):
    """Returns the rendered error dialog"""
    c.dialog_title = header or _('Default Header')
    if dyntext and text:
        c.dialog_text = text % dyntext
    else:
        c.dialog_text = text or _('Default Text')
    c.url_ok = url_ok
    return render('/dialogs/error_%s.mako' % get_dialog_template(ctx))


def confirm(header=None, text=None, url_ok=None, url_abort=None, dynfunc=None):
    """Decorator to be used for actions that need to be confirmed by the user"""

    def validate(func, self, *args, **kwargs):
        """Checks if the confirmed paramenter is set to 1 and call the origin
        function. If the call is not confirmed, then render a confirm dialog"""
        if DIALOGS_ENABLED:
            # Check if the url include a parameter 'confirmed' which is set to
            # '1'
            url_dict = request.environ.get('pylons.routes_dict')
            if url_dict.get('confirmed') == '1':
                log.debug('calling function %s is confirmed. Proceeding.'
                          % func.__name__)
                return func(self, *args, **kwargs)
            else:  # show dialog
                # Build url for ok
                url_ok = url(controller=url_dict.get('controller'),
                             action=url_dict.get('action'),
                             id=url_dict.get('id'),
                             confirmed=1)
                log.debug('calling function %s is NOT confirmed. Redirecting.'
                          % func.__name__)
                # Save request.params in a session. Otherwise they will get
                # lost after the user confirms the action (new request).
                # session['request.params'] = request.params
                # session.save()

                # Use function to include dynamic content to the confirm
                # dialog, function will be called with the id of the origin
                # call
                dyntext = None
                if dynfunc:
                    dyntext = dynfunc(args[0])
                return render_confirm(_(header), _(text), url_ok, url_abort,
                                      dyntext)
        else:  # No dialogs at all
            return func(self, *args, **kwargs)

    return decorator(validate)


def success(header=None, text=None, url_ok=None, dynfunc=None):
    """Simple function which can be called after an action was successfully
    done. It returns a rendered success-dialog"""
    url_ok = url_ok or url(controller="case", action="digest")
    return render_success(header, text, url_ok, dynfunc)


def error(header=None, text=None, url_ok=None, dynfunc=None):
    """Simple function which can be called after an action was  done with
    errors. It returns a rendered error-dialog"""
    url_ok = url_ok or url(controller="case", action="digest")
    return render_error(header, text, url_ok, dynfunc)


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8:
