# -*- coding: utf-8 -*-
# Copyright 2007, 2008, 2010 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

"""Helper functions

Consists of functions to typically be used within templates, but also
available to Controllers. This module is available to both as 'h'.
"""

import locale
import logging
import re
import os
import datetime
import warnings

from ConfigParser import NoOptionError
from functools import wraps

from libmpuls.service.tags import TagSetterConfig, TagSetter
from pylons import config, app_globals as g, request, url
from mpulsweb.lib.translation import _

# Use url_for from prior versions
from routes import url_for

from webhelpers.html import escape, literal, url_escape
from webhelpers.html.tags import *

from formed.meta import UNKNOWN_DATE
from formed.meta.data import RepeatNode, GroupNode
from formed.meta.document import containing_repeat_group
from formed.renderer.helpers import tag, folding_helper_functions, folding_init

from mpulsweb.lib.security import hasRole, getKAName, \
    get_db_parameters, getDbName, current_csrf_token

from mpulsweb.lib.helper.filters import shorten, NA
from psycopg2.extensions import adapt as psycoadapt
from mpulsweb.lib.db import db

# normally we would import session from mpulsweb.lib.base, but to avoid
# circular imports we import it from app_globals.
from mpulsweb.lib.app_globals import session


log = logging.getLogger(__name__)


def csrf_token():
    """Return HTML code for CSRF protection in HTML forms if enabled.

    If CSRF protection is enabled in g.mpuls_config, the return value is
    an HTML-fragment with a hidden input field containing the CSRF
    token. Otherwise it is an empty string.

    In both cases the return value is wrapped in
    webhelpers.html.literal. In the first case because the string is
    already HTML and must not be escaped again by the templating engine,
    in the latter case because literal value behave differently than
    plain strings and the behavior of the program should not change in
    subtle ways if CSRF protection is disabled. E.g. when concatenating
    a literal and a unicode string the unicode string is html-escaped
    and the result is a literal.
    """
    if g.mpuls_config.is_enabled('security', 'require-csrf-tokens'):
        return literal('<input type="hidden" name="csrftoken" value="%s">'
                       % (escape(current_csrf_token()),))
    return literal("")


# unicode helper
def ensure_unicode(s):
    """Convert s to a unicode object.
    If it's alredy a unicode object, s is returned as is.  Otherwise
    it's converted to a byte-string which is assumed to be in utf-8 and
    then converted to unicode."""
    if isinstance(s, unicode):
        return s
    return unicode(str(s), 'utf-8')


def render_logbook_typeoptions(item, selection=()):
    out = []
    for i in item.get('options', []):
        if isinstance(i, dict):
            out.append('<optgroup label="%s">' % i.get('label'))
            out.extend(render_logbook_typeoptions(i, selection))
            out.append('</optgroup>')
        else:
            key, label = i
            if int(key) in selection:
                selected = "selected"
            else:
                selected = ""
            out.append('<option value="%s" %s>%s</option>'
                       % (key, selected, label))
    return out


def convertForDigest(value):
    """Will convert given values to it's string representation for correct
    display in digest. Will localize dates and shown and return string
    'unknown' if it is the unknown value"""
    if isinstance(value, datetime.date):
        if value == datetime.date(1, 1, 1):
            return _('unknown')
        else:
            return format_date(value)
    if isinstance(value, int):
        if value == -9999999:
            return _('unknown')
    if value is None:
        return ""
    return value

# Helper functions used in the UI to show some information like status messages
# or icons


def get_sorted_db_from_selectionlist(filter=""):
    """Returns a sorted list of available dbs. dbnames are useually defined as
    a name followed by a number (i.e inte00010). Sorting will be done on the
    numberpart. If dbname does not contain any number the assume number '0'
    which results in no sorting. """
    d = get_db_selectionlist(filter)
    s = {}
    name = re.compile('\D*')
    for k in d.keys():
        num = name.split(k)[-1] or "0"
        i = int(num)
        s[i] = k

    l = []
    for id in sorted(s.keys()):
        l.append(s.get(id))
    return l


def get_db_selectionlist(filter=""):
    """
    Returns a dictionary used for listing available db on the loginscreen.
    """
    dblist = g.mpuls_dblist.get_list()
    if not filter:
        return dblist
    filtered_dblist = {}
    for k in dblist.keys():
        if not filter.match(k):
            continue
        filtered_dblist[k] = dblist.get(k)
    return filtered_dblist


def status_message(s):
    return "%s: %s" % (format_datetime(datetime.datetime.now()), s)


def get_logouttime():
    """Estimate log-out time and return it as a formatted string (HH:MM).
    The estimated logout time is two minutes before the time the session
    actually expires to make sure it's earlier.
    """
    timeout = int(config.get("beaker.session.timeout"))
    return format_time(datetime.datetime.now()
                       + datetime.timedelta(seconds=timeout - 120), True)


def get_countdowntime():
    """Get countdown time (in seconds) derived from session timeout.
    The countdown time is two minutes before the time the session
    actually expires to make sure it's earlier.
    """
    timeout = int(config.get("beaker.session.timeout"))
    return timeout - 120


def get_TagSetter(user, config):
    # TODO:
    # 1. Use established db-connection
    # 2. Do create (and parse config) a TagSetter every time.
    tagconfig = TagSetterConfig(g.taggingConfFile)
    tagconfig.set_database(db)
    return TagSetter(tagconfig)


def getHelp(help_id):
    '''
    Returns a HTML link with a help icon if help is present for help_id. Else
    return empty string
    '''
    out = []
    help = g.helpData.getHelp(help_id)
    if help:
        out.append(tag("a", href=url(controller='annotations',
                                     action='help', id=help_id),
                       target="_blank"))
        out.append(tag("img", src=url('/images/icons/formular/help.png'),
                       alt="help"))
        out.append('</a>')
    return literal("".join(out))


def get_app_title():
    return config.get('mpuls.app.title')


def get_num_of_other_datasets():
    n = 0
    apps = re.compile('app:')
    for sec in g.mpuls_ini.sections():
        if apps.match(sec):
            n += 1
    return n


def get_urls_for_other_datasets():
    apps = re.compile('app:')
    l = []
    for sec in g.mpuls_ini.sections():
        if apps.match(sec):
            try:
                name = g.mpuls_ini.get(sec, 'mpuls.app.name')
                url2 = g.mpuls_ini.get(sec, 'mpuls.app.url')
                if request.url.find(url2) < 0:  # current url is not switch-url
                    l.append((name, url2, False))
            except:
                pass
    # Administration is only needed if there are more than one datasets
    if get_num_of_other_datasets() > 1:
        try:
            url = g.mpuls_ini.get('app:redirector', 'location')
            if request.url.find(url) > 0 and hasRole(['admin']):
                l.append(
                    (_('Userpermissions'), os.path.join(url, 'permissions'),
                     True))
        except:
            pass
    return l


def get_mpuls_app_infos():
    apps = {}
    apps_re = re.compile('app:(?P<appname>.+)')
    for sec in g.mpuls_ini.sections():
        match = apps_re.match(sec)
        if match:
            try:
                name = g.mpuls_ini.get(sec, 'mpuls.app.name')
            except NoOptionError:
                continue
            url = g.mpuls_ini.get(sec, 'mpuls.app.url')
            other_url = g.mpuls_ini.get(sec, 'mpuls.app.other_url', None)
            appname = match.group("appname")
            app = MpulsAppInfo(appname, name, url, other_url)
            apps[appname] = app
    return apps


class MpulsAppInfo(object):

    def __init__(self, appname, name, url, other_url=None):
        self.appname = appname
        self.name = ensure_unicode(name)
        self.url = url
        self.other_url = other_url

    def get_name(self):
        return self.name

    def get_url(self):
        return self.url

    def get_appname(self):
        return appname

    def get_permissions_url(self):
        return "%s/permissions" % self.other_url


# Helper functions to get some information about the logged in user

def getRole():
    try:
        user = session['USER_AUTHORIZED']
        return user.getRole()
    except KeyError:
        return ''


def getUserLastname():
    try:
        return session['USER_AUTHORIZED'].last_name
    except KeyError:
        return ''

def getUserFirstname():
    try:
        return session['USER_AUTHORIZED'].first_name
    except KeyError:
        return ''


def getUserFullname():
    try:
        user = session['USER_AUTHORIZED']
        return "%s, %s" % (user.last_name, user.first_name)
    except KeyError:
        return ''


def getLogin():
    try:
        user = session['USER_AUTHORIZED']
        return user.login
    except KeyError:
        return ''


def getUserId():
    try:
        user = session['USER_AUTHORIZED']
        return user.id
    except KeyError:
        return ''


def get_last_failed_login():
    user = session.get("USER_AUTHORIZED")
    if user is None:
        return None
    return g.login_db.get_failed_login(user.full_login)


def secure_sql_string(value):
    # HACK! Converting unicode to other encoding only works after the
    # connection of the db-api has been prepared. Maybe this be removed in
    # newer version of psycopg.
    # See
    # http://www.postgresql.org/message-id/AANLkTikf0YTUwo7vgRVyoQtzspOmZh0q42XLdqOtdK62@mail.gmail.com
    adapted = psycoadapt(value)
    adapted.prepare(db.getConnection())
    quoted = adapted.getquoted()
    # quoted is a bytestring here. As we rely on unicode further. ensure
    # we return unicode
    return ensure_unicode(quoted)

# Format functions


def format_number(number, precision=2, use_precision_for_integer=False):
    """
    Format a numeric value based on current locale an precision

    For integer values (integer here means values without floating point) will
    get a precision of 0 per default unless use_precision_for_integer is set to
    True.
    """
    if is_integer(number) and not use_precision_for_integer:
        precision = 0
    fmt = "%%.%sf" % precision
    retval = locale.format(fmt, (number), 1)
    return retval


def is_integer(number):
    if number == float(int(number)):
        return True
    return False


def format_date(dt):
    if dt is None:
        return ''
    if dt == UNKNOWN_DATE:
        return _("unknown")
    try:
        cs = locale.nl_langinfo(locale.CODESET)
        return unicode(dt.strftime(locale.nl_langinfo(locale.D_FMT)), cs)
    except Exception, err:
        log.error(err)
        return dd_mm_YYYY(dt)


def format_time(dt, approximate=False):
    """
    Format time with the correct locale. If the optional parameter "approximate"
    is set, the minutes will be rounded to the previous multiple of 5
    """
    if dt is None:
        return ''
    try:
        cs = locale.nl_langinfo(locale.CODESET)
        if approximate:
            dt = dt.replace(minute=int(5 * (dt.minute / 5)))
        return unicode(dt.strftime('%H:%M'), cs)
    except Exception, err:
        log.error(err)
        return ''


def format_datetime(dt):
    if dt is None:
        return ''
    if dt == UNKNOWN_DATE:
        return _("unknown")
    try:
        cs = locale.nl_langinfo(locale.CODESET)
        return unicode(dt.strftime(locale.nl_langinfo(locale.D_T_FMT)), cs)
    except Exception, err:
        log.error(err)
        return dd_mm_yyyy_HH_MM(dt)


def timedelta_in_minutes(ti):
    if ti is None:
        return None
    return round(float(ti.days * (24 * 60)) + ti.seconds / 60.0)


def timedelta_in_weeks(ti, empty=None):
    if ti is None:
        return empty
    return format_number(float(ti.days) / 7.0
                         + ti.seconds / (60.0 * 24.0 * 7.0))


def timedelta_in_years(ti, empty=None, precision=None):
    if ti is None:
        return empty
    return format_number(float(ti.days) / 365.0
                         + ti.seconds / (60.0 * 24.0 * 365.0))


def get_age(ti, empty=None):
    if ti is None:
        return empty
    return int(float(ti.days) / 365.0
               + ti.seconds / (60.0 * 24.0 * 365.0))


def human_timedelta_days(td, empty=None):
    if td is None:
        return empty

    sign = ""
    if td < datetime.timedelta(0):
        sign = "-"
        td = abs(td)

    days = "Tag" if td.days == 1 else "Tage"
    return "%s%d %s" % (sign, td.days, days)


def human_timedelta(td, empty=None):
    if td is None:
        return empty

    sign = "-" if td < datetime.timedelta(0) else ""

    # The representation of negative time deltas is a bit surprising
    # (td.days is negative, but td.seconds is positive), so we work with
    # the absolute value for the rest of the conversion
    td = abs(td)

    # Only mention days if the delta is at least a day
    days = human_timedelta_days(td) + ", " if td.days > 0 else ""

    mins, secs = divmod(td.seconds, 60)
    hours, mins = divmod(mins, 60)
    return "%s%s%d:%02d:%02d" % (sign, days, hours, mins, secs)


def format_user_name_short(user):
    """Return the user's name in a shortened form: 'LASTNAME, INITIAL',
    where INITIAL is the first character of the first name. Each
    component is replaced in the output by '-/-' if it is missing.
    """
    return "%s, %s" % (NA(user.last_name), NA(user.first_name[:1]))


# formatters for the format_object funtion.
format_object_handlers = [
    (datetime.date, format_date),
    (datetime.datetime, format_datetime),
    (datetime.time, format_time),
    (type(None), lambda obj: ''),
]


def format_object(obj):
    """Return a string representation of obj suitable for the user.
    Date/time objects are formatted with the appropriate format_date
    function.  None objects are converted to an empty string.  Objects
    that are already strings (bytestrings or unicode) are returned as
    is.  Other objects are converted using str().
    """
    for typeobj, handler in format_object_handlers:
        if isinstance(obj, typeobj):
            return handler(obj)
    if isinstance(obj, basestring):
        return obj
    return str(obj)


# Fallback functions for the localized format_* functions. If localisation
# fails because of dates before 1900 (which aren't meaningfull anyway') the
# fallback functions are called. These functions will convert dates and time in
# german date and time formats

def dd_mm_YYYY(d, empty=''):
    """Format the date of the date/time object d as 'dd.mm.YYYY'.
    The parameter d may be a string or a datetime instance.  If d is
    None, the function return the optional parameter empty (default "").
    """
    # datetime.strftime() dislikes years before 1900.
    if d is None:
        return empty
    # TODO: what should happen if d is a unicode object?
    if type(d) == str:
        try:
            y, m, d = d.split('-')
            d = datetime.date(int(y), int(m), int(d))
        except:
            return empty
    return "%02d.%02d.%d" % (d.day, d.month, d.year)


def HH_MM(d):
    """Format the time object d as 'HH:MM'.
    The parameter d may be a time, datetime or date instance.  If it is
    a date instance, time 00:00 is assumed.  If d is None, the function
    returns the empty string.
    """
    # datetime.strftime() dislikes years before 1900.
    if d is None:
        return ''
    if type(d) is datetime.date:
        # not so smart
        d = datetime.datetime(d.year, d.month, d.day)
    return "%02d:%02d" % (d.hour, d.minute)


def dd_mm_yyyy_HH_MM(d, empty=''):
    """Format the time/date object d as 'dd.mm.YYYY HH:MM'.
    The parameter d may be a time, datetime or date instance.  If it is
    a date instance, time 00:00 is assumed.  If d is None, the function
    return the optional parameter empty (default"").
    """
    if d is None:
        return empty
    return dd_mm_YYYY(d) + " " + HH_MM(d)


def get_formular_url(formed, id, name, page=None, action_arg='select'):
    # id is like fieldname:rgid:caseid
    # The handling of links for Element in RG differs depending if they are
    # located within a outline RG or not
    # RG Types:
    # 0 = not in rg at all
    # 1 = within embedded rg
    # 2 = within outline rg
    rg_type = formed.name2rgtype.get(name, 0)
    if page is None:
        page = formed.name2page[name].getName()
    if rg_type in (0, 1):
        # Ignore rgid, as it is not needed in this case.
        element_id = ":".join(id.split(':')[1:][-1:])
    else:
        # NOTE: Nested outline RG are not supported.
        element_id = ":".join(id.split(':')[1:][-2:])
    anchor_id = ":".join(id.split(':')[1:])
    return url(controller='formularpage', action=action_arg,
               id=('%s:%s' % (page, element_id)),
               anchor=('f_%s:%s' % (name,
                                  anchor_id)))


def get_url_for_nc(nc_name, case_id):
    nc = g.formedTree.name2nc.get(nc_name)
    if nc is None:
        return None
    rg_nc = containing_repeat_group(nc)
    if rg_nc:
        # For fields in repeat groups we cannot link directly to the
        # field, because we don't know which instance we should link to.
        #
        # For fields in inline RGs we link to the parent of the RG which
        # we assume to be in master_tbl. For standalone RGs we link to
        # the corresponding overview page.
        rg_type = g.formedTree.name2rgtype[nc_name]
        if rg_type == 1:
            # inline RG
            nc_name = rg_nc.getParent().getName()
        elif rg_type == 2:
            # standalon RG, get_formular_url will not work correctly
            # here, so we handle this case explicitly here
            return url(controller="formularpage", action="select",
                       id=('%s:%s' % (rg_nc.getName(), case_id)))
    urlid = ":".join((nc_name, str(case_id)))
    return get_formular_url(g.formedTree, urlid, nc_name)


def deprecated(message):
    """
    A decorator to raise a warning message when using the decorated function or
    method.

    Usage:
        @depracted("myfunc will be dropped with version 3.2. Please use"
                   "myfunc2 instead")
        def myfunc(...):
            ...
    """
    def wrapper(f):
        @wraps(f)
        def inner(*args, **kwargs):
            warnings.warn(message, category=DeprecationWarning,  stacklevel=2)
            return f(*args, **kwargs)
        return inner
    return wrapper


def get_field_identifier(id, name, it=None, idx=0):
    """Will return the identifier of a field. If the field in in a rg the
    identifier of the first occourance will be returned on default."""
    # Map generic fieldnames to specific fieldnames
    # get name of the field
    nc = g.formedTree.findByName(name)
    if not nc:
        return None

    # is the field within a rg. if so get name
    parent = nc.getParent()
    rg = None
    while parent:
        if isinstance(parent, RepeatNode):
            rg = parent
            break
        parent = parent.getParent()

    if not rg:
        return "%s:%d" % (name, id)

    # get instance of the rg
    rgc = it.getItem("%s:%d" % (rg.getName(), id))
    if not rgc:
        return None
    rgs = rgc.getChildren()
    if not rgs:
        log.debug("No repeat group found for %r: create new", name)
        rg = rgc.createInstance()
    else:
        rg = rgs[idx]
    return '%s:%s:%s' % (name, rg.getInternalIdentifier(), id)


def format_boolean(value):
    return u"Ja" if value else "Nein"


def can_use_evaluation(evaluation):
    """Checks whether the current user can use the given evaluation.

    The evaluation parameter is expected to be one of the dictionaries
    defining an evaluation in the JSON config file, e.g.

         {"name": "Logbuch", "id": "4", "config": "eval_logbook.xml",
          "roles": ["cm"]}

    The "roles" member is taken to be a list of the roles which may use
    the evaluation and hasRole is used to check whether the current use
    has one of those roles. If it's missing, all roles may use the
    evaluation.
    """
    roles = evaluation.get("roles")
    return roles is None or hasRole(roles)


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
