#! /usr/bin/env python
# -*- coding: utf-8 -*-

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Remove when not longer using python 2.5
from __future__ import with_statement

import json
from StringIO import StringIO
import openpyxl
from datetime import datetime, date
import os
import decimal

from optparse import OptionParser

import psycopg2
import psycopg2.extras as extras
import psycopg2.extensions

import smtplib
from email import encoders
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText

# make psycopg2 return unicode strings in order to handle non-ASCII
# characters like äöü properly.
psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)

### Constants to configure
def load_config(config_file_name):
    with open(config_file_name) as config_file:
        config = json.load(config_file)
    map(lambda x:set_global_variable(x[0], x[1]), config.iteritems())


def set_global_variable(key, value):
    globals()[key] = value


#
# Helpers for pyExcelerator
#

date_format = "DD.MM.YYYY"

def write_header_to_sheet(header, sheet):
    '''Write the header to the sheet'''
    for column, name in enumerate(header, start=1):
        sheet.cell(row=1, column=column, value=name)

def write_rows_to_sheet(rows, sheet):
    '''Write the rows to the sheet'''
    for row_num, row in enumerate(rows, start=1):
        row_num += 1 # Skip header row
        for col_num, value in enumerate(row, start=1):
            value, number_format = encode(value)
            cell = sheet.cell(row=row_num, column=col_num, value=value)
            cell.number_format = number_format

def encode(value):
    '''Encode a value to use with pyExcelerator'''
    number_format='General'
    if value is None:
        return ("", number_format)
    if isinstance(value, datetime):
        number_format = date_format
        return (value, number_format)
    if isinstance(value, date):
        number_format = date_format
        value = datetime(value.year, value.month, value.day)
        return (value, number_format)
    if isinstance(value, decimal.Decimal):
        return(float(value), number_format)
    if isinstance(value, basestring) and len(value) > 32765:
        value = value[:32765]
    if not isinstance(value, (basestring, int, long, float)):
        return (str(value), number_format)
    return (value, number_format)


#
# Parse user options
#
parser = OptionParser()
parser.add_option("--year",
                  action="store",
                  dest="year",
                  default=date.today().year,
                  help="Year to filter the reports. Default: current year")
parser.add_option("--dont-send-email",
                  dest="send_email",
                  action="store_false",
                  default=True,
                  help=("Don't send emails with the reports but write them"
                        " to a file instead 'tmp-email-$TRAEGER.txt'"))
parser.add_option("--keep-files",
                  action="store_true",
                  dest="keep_files",
                  default=False,
                  help="Don't delete generated files")
parser.add_option("--force-recipient",
                  action="store",
                  dest="forced_recipient",
                  default=None,
                  help=("Send mails only to this address for"
                        " for easier testing"))
parser.add_option("--config",
                  action="store",
                  dest="config_file",
                  default=None,
                  help="Configuraton file - mandatory")

(options, args) = parser.parse_args()


if not options.config_file:
    print("Configuration file missing. Aborting.")
    exit(2)
else:
    load_config(options.config_file)
    


os.chdir('/tmp')

#
# Process each Traeger
#

print "\n** Start %s" % datetime.now()

connection = psycopg2.connect("dbname='ka_stservicestelleauswertung102_db' port='5437'")
cur = connection.cursor()
for traeger in traeger_mapper:

    print "**** Processing: %s" % traeger
    workbook = openpyxl.Workbook()
    ws = workbook.active
    ws.title = 'Metadaten'
    ws.cell(row=1, column=1, value='Auswertungen JMD Strukturteil')
    ws.cell(row=2, column=1, value='Datum')
    cell=ws.cell(row=2, column=2, value=date.today())
    cell.number_format=date_format

    # Process all queries and write the results to disk
    for query in query_list:
        print query['name']
        is_header = True
        sheet = workbook.create_sheet('%s' % query['name'])

        cur.execute('''SELECT fkz,
                              database,
                              name
                       FROM fkz_tbl
                       WHERE bundesland_id >= 0 AND
                             fkz::integer/1000 = %s
                             AND name NOT LIKE '%%(bis%%'
                       ORDER BY database;''',
                    (traeger_mapper[traeger]['case_nr'],))
        rows= []
        for db in cur:
            cur2 = connection.cursor()

            try:
                cur2.execute(query['query'], ({'year': options.year, 'fkz': db[0]}))
            except psycopg2.ProgrammingError, e:
                print ("ProgrammingError while executing query."
                       " Details follow: %s" % e)
                continue
            except psycopg2.InternalError, e:
                print ("InternalError while executing query."
                       " Details follow: %s" % e)
                continue

            # Append header row
            if is_header:
                colnames = ['FKZ', 'JMD'] + [i[0] for i in cur2.description]
                write_header_to_sheet(colnames, sheet)
                is_header = False

            # Append evaluation results
            for result in cur2:
                rows.append([db[0], db[2]] + [i for i in result])

        write_rows_to_sheet(rows, sheet)
    workbook.save("%s.xlsx" % traeger)       
#
# Attach the evaluation CSV files to emails and send them to the Bundestutoren
# See http://docs.python.org/2/library/email-examples.html for details
#

text_template = u"""\
Anbei erhalten Sie die Übersichtsauswertung JMD Strukturteil
für das Jahr %(year)s mit Stand %(date)s als xlsx-Datei.

Viele Grüße
Ihr mpuls-Team
"""

for traeger in traeger_mapper:
    outer = MIMEMultipart()
    outer['Subject'] = 'Auswertungen JMD Strukturteil'
    outer['To'] = ', '.join(traeger_mapper[traeger]['emails'])
    outer['From'] = 'mpuls@support.intevation.de'

    outer.attach(MIMEText(text_template
                          % {"year": options.year,
                             "date": date.today().strftime("%d.%m.%Y")},
                          _charset="utf8"))

    filename = "%s.xlsx" % traeger
    with open(filename) as fp:
        msg = MIMEBase('application',
                       'vnd.openxmlformats-officedocument.spreadsheetml.sheet')
        msg.set_payload(fp.read())
    encoders.encode_base64(msg)

    msg.add_header('Content-Disposition', 'attachment', filename=filename)
    outer.attach(msg)

    # Delete the files
    if not options.keep_files:
        os.remove(filename)

    composed = outer.as_string()
    # Write to disk or send per email
    if options.send_email:
        s = smtplib.SMTP('localhost')
        recipients = traeger_mapper[traeger]['emails']
        if options.forced_recipient is not None:
            recipients = [options.forced_recipient]
        s.sendmail('mpuls@support.intevation.de', recipients, composed)
        s.quit()
    else:
        with open('tmp-email-%s.txt' % traeger, 'w') as fp:
            fp.write(composed)

print "** End %s" % datetime.now()

