# -*- coding: utf-8 -*-
#
# Copyright (c) 2008 by Intevation GmbH
#
# FormEd is free software under the terms of the
# GNU General Public License Version 3. See file
# LICENSE coming with the source of FormEd for details.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
# Frank Koormann <frank.koormann@intevation.de>
# Björn Ricks <bjoern.ricks@intevation.de>
#

import sys
from cgi import escape

import formed.meta.data as data
from formed.meta.nodecomponents import Node, Leaf


class Type:

    def __init__(self, name, nc):
        self.name = name
        self.nc = nc

    def buildDescriptionChain(self):
        n = self.nc
        descs = []
        while n:
            desc = n.getDescription()
            if desc:
                if descs and descs[0] == desc:
                    n = n.getParent()
                    continue
                descs.insert(0, desc)
            n = n.getParent()
        return descs

    def dumpAnnotation(self, out):
        pass

    def dump(self, out):
        pass


class BasedType(Type):

    def __init__(self, name, nc, base):
        Type.__init__(self, name, nc)
        self.base = base

    def dump(self, out):
        pass


class TextAreaType(Type):

    def __init__(self, name, nc):
        Type.__init__(self, name, nc)

    def dump(self, out):
        pass


class TextType(Type):

    def __init__(self, name, nc):
        Type.__init__(self, name, nc)

    def dump(self, out):
        pass


class DateType(Type):

    def __init__(self, name, nc):
        Type.__init__(self, name, nc)

    def dump(self, out):
        pass


class IntegerType(Type):

    def __init__(self, name, nc):
        Type.__init__(self, name, nc)

    def dump(self, out):
        pass


class EnumerationType:

    def __init__(self, name, description, keys_values):
        self.name = name
        self.keys_values = keys_values
        self.references = [(name, description)]

    def isSame(self, other):
        return self.keys_values == other.keys_values

    def dump(self, book):
        sheet = book.addSheet(self.name)
        row = 1
        for k, v in self.keys_values:
            sheet.cell(row=row, column=1, value='%d' % k)
            sheet.cell(row=row, column=2, value='%s' % escape(v))
            row +=1
        row+=2
        sheet.cell(row=row, column=1, value='Bezug:')
        row +=1
        for k, v in self.references:
            sheet.cell(row=row, column=1, value='%s' % escape(k))
            sheet.cell(row=row, column=2, value='%s' % escape(v))
            row +=1

    def addReference(self, name, description):
        self.references.append((name, description))


class TableType:

    def __init__(self, name, maxOccurs="unbounded"):
        self.name = name
        self.items = []
        self.maxOccurs = maxOccurs

    def dump(self, out):
        pass

    def dumpTypes(self, out):
        for item in self.items:
            item.dump(out)

    def appendItem(self, item):
        self.items.append(item)


class CaseType:

    def __init__(self):
        self.tables = []

    def dumpTables(self, out):
        for table in self.tables:
            table.dump(out)
            table.dumpTypes(out)

    def appendTable(self, table):
        self.tables.append(table)


class SchemaCreator:

    def __init__(self):
       pass

    def createSchema(self, root):
       self.enums = []
       case = CaseType()
       master = TableType("master", "1")
       case.appendTable(master)
       self.case = case
       self.tableStack = [master]
       self._createRecursive(root)

    def getCase(self):
        return self.case

    def findEnumeration(self, enum):
        for e in self.enums:
            if e.isSame(enum):
                return e
        return None

    def dumpEnumerations(self, out):
        for e in self.enums:
            e.dump(out)

    def _createRecursive(self, node):
        for child in node.children:
            self._createChild(child)

    def _createDate(self, date):
        self.tableStack[-1].appendItem(DateType(date.getName(), date))

    def _createBool(self, bool):
        key_values = [
            (-1, 'Keine Angabe'),
            (0, 'Nein'),
            (1, 'Ja')]

        enum_name = "ja_nein"
        enum_description = "Ja/Nein"

        enum = EnumerationType(enum_name, enum_description, key_values)

        base = self.findEnumeration(enum)

        if not base:
            base = enum
            self.enums.append(enum)

        self.tableStack[-1].appendItem(BasedType(bool.getName(), bool,
                                                 base.name))

    def _createText(self, text):
        self.tableStack[-1].appendItem(TextType(text.getName(), text))

    def _createTextArea(self, text):
        self.tableStack[-1].appendItem(TextAreaType(text.getName(), text))

    def _createInteger(self, integer):
        self.tableStack[-1].appendItem(IntegerType(integer.getName(), integer))

    def _createChoice(self, choice):
        if choice.getMultiple():
            print >> sys.stderr, "WARNING: No support for multiple choices."
            return

        kvs = {}
        stack = choice.children[:]
        try:
            while stack:
                c = stack.pop()
                if isinstance(c, data.ExternalChoiceListLeaf):
                    cc = c.getChildren()
                    if cc:
                        stack.extend(cc)
                elif isinstance(c, data.GroupNode):
                    stack.extend(c.children)
                # Also add options if they are defined in a conditional
                elif isinstance(c, data.ConditionalNode):
                    stack.extend(c.children)
                elif isinstance(c, data.BoolLeaf):
                    desc = c.getDescription()
                    if not desc:
                        desc = ""
                    kvs[int(c.getValue())] = desc
        except ValueError:
            print >> sys.stderr, "WARNING: Choice '%s' contains no integer values" % name
            return

        if not kvs:
            print >> sys.stderr, "WARNING: choice '%s' contains no values" % choice.getName()
            return

        keys_values = list(kvs.iteritems())
        keys_values.sort()

        enum_name = choice.getName()
        enum_description = choice.getDescription()

        enum = EnumerationType(enum_name, enum_description, keys_values)

        base = self.findEnumeration(enum)

        if not base:
            base = enum
            self.enums.append(enum)
        else:
            base.addReference(enum_name, enum_description)

        self.tableStack[-1].appendItem(BasedType(choice.getName(), choice,
                                                 base.name))

    def _createRadioGroup(self, radio):
        kvs = {}
        stack = radio.children[:]
        while stack:
            c = stack.pop()
            if isinstance(c, data.BoolLeaf):
                desc = c.getDescription()
                if not desc:
                    desc = ""
                kvs[int(c.getValue())] = desc

        if not kvs:
            print >> sys.stderr, "WARNING: radio group '%s' contains no values" % radio.getName()
            return

        keys_values = list(kvs.iteritems())
        keys_values.sort()
        enum_name = radio.getName()
        enum_description = radio.getDescription()
        enum = EnumerationType(enum_name, enum_description, keys_values)

        base = enum
        self.enums.append(enum)

        self.tableStack[-1].appendItem(BasedType(radio.getName(), radio,
                                                 base.name))

    def _createGroup(self, group):
        repeat = group.isRepeat()
        if repeat:
            table = TableType(group.getName().replace("rg_", "", 1))
            self.tableStack.append(table)
            self.case.appendTable(table)

        for child in group.children:
            self._createChild(child)

        if repeat:
            self.tableStack.pop()

    def _createRepeat(self, repeat):
        table = TableType(repeat.getName().replace("rg_", "", 1))
        self.tableStack.append(table)
        self.case.appendTable(table)

        for child in repeat.children:
            self._createChild(child)

        self.tableStack.pop()

    def _createChild(self, child):
        if isinstance(child, Leaf):
            if isinstance(child, data.IntLeaf):
                self._createInteger(child)
            elif isinstance(child, data.DateLeaf):
                self._createDate(child)
            elif isinstance(child, data.TextLeaf):
                self._createText(child)
            elif isinstance(child, data.TextAreaLeaf):
                self._createTextArea(child)
            elif isinstance(child, data.BoolLeaf):
                self._createBool(child)
        elif isinstance(child, Node):
            if isinstance(child, data.GroupNode):
                self._createGroup(child)
            elif isinstance(child, data.RepeatNode):
                self._createRepeat(child)
            elif isinstance(child, data.RadioNode):
                self._createRadioGroup(child)
            elif isinstance(child, data.ChoiceNode):
                self._createChoice(child)
            else:
                self._createRecursive(child)


class Book:

    def __init__(self):
        from openpyxl import Workbook
        self.workbook = Workbook()
        # Delete the empty sheet created by default.
        del self.workbook[self.workbook.active.title] 

    def addSheet(self, title):
        return self.workbook.create_sheet(title[:31])

    def writeXLS(self, fname):
        self.workbook.save(fname)


class ExportKeyValueTableAsXLS:

    def __init__(self, document):
        self.root = document.root

    @staticmethod
    def getDescription():
        return "Export Key-Value pairs as XLS."

    def doExport(self, path):
        creator = SchemaCreator()
        creator.createSchema(self.root)
        out = Book()
        case = creator.getCase()
        case.dumpTables(out)
        creator.dumpEnumerations(out)
        out.writeXLS(path)

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
