# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3.
# See LICENSE comming with the source of 'mpuls'
# for details.
#
# author: Torsten Irländer <torsten.irlaender@intevation.de>

import logging

try:
    import xml.etree.ElementTree as etree
except ImportError:
    import elementtree.ElementTree as etree

from libmpuls.db.db import get_db

log = logging.getLogger(__name__)

ADD_TAG_SQL = """SELECT create_nm_tag_case_ds(%(tagid)s, %(caseid)s)"""
DEL_TAG_SQL = """SELECT delete_nm_tag_case_ds(%(tagid)s, %(caseid)s)"""


class TagSetterConfig(object):

    def __init__(self, filename, username=None, password=None,
                 database=None, host=None, port=None, caseid=None):
        self.dbname = database
        self.username = username
        self.password = password
        self.host = host
        self.port = port
        self.caseid = caseid  # None means all cases
        self._database = None
        try:
            xml = open(filename, 'r')
            self.xml_doc = etree.parse(xml)
            xml.close()
        except StandardError, err:
            log.error('Could not load xml tagsetter config: %s' % err)
            raise

    def set_db(self, username, password, database, host=None, port=None):
        # FIXME: this method should be removed.
        self.dbname = database
        self.username = username
        self.password = password
        self.host = host
        self.port = port

    def set_database(self, db):
        self._database = db

    def get_database(self):
        if not self._database:
            self._database = get_db(self.username, self.password,
                                    self.dbname, self.host, self.port)

        return self._database

    def _get_master_tbl(self):
        if self.caseid:
            return ("SELECT * FROM master_tbl_view WHERE id IN (%s)"
                    % self.caseid)
        else:
            return "SELECT * FROM master_tbl_view"

    def _get_check(self, id):
        for eval in self.xml_doc.findall('check'):
            tagid = eval.attrib.get('tagid')
            if str(id) == tagid:
                return eval

    def get_checks(self):
        """Returns a list of ids of available checks.

        @return: list of ids of available checks
        @rtype: List of integers
        """
        ids = []
        for eval in self.xml_doc.findall('check'):
            tagid = eval.attrib.get('tagid')
            if tagid:
                ids.append(tagid)
        return ids

    def get_name(self, id):
        config = self._get_check(id)
        return config.findtext('name')

    def get_desc(self, id):
        config = self._get_check(id)
        return config.findtext('desc')

    def get_addsql(self, id):
        config = self._get_check(id)
        sqlq = []
        table = self._get_master_tbl()
        for sql in config.findall('addtag'):
            sqlq.append(sql.findtext('sql') % {'table': table})
        return sqlq

    def get_delsql(self, id):
        config = self._get_check(id)
        sqlq = []
        table = self._get_master_tbl()
        for sql in config.findall('deltag'):
            sqlq.append(sql.findtext('sql') % {'table': table})
        return sqlq


class TagSetter(object):

    def __init__(self, config):
        self.config = config
        self.checks = self.config.get_checks()

    def perform(self, caseid=None):
        """Check cases and set tags.
        If caseid is provided the caseid in the TagSetterConfig will be
        overwritten.
        """
        if caseid:
            # Overwrite configcase id
            self.config.caseid = caseid
        db = self.config.get_database()
        conn = db.getConnection()
        ca = 0
        cd = 0
        for tagid in self.checks:
            cd += self.del_tag(tagid, conn)
            ca += self.add_tag(tagid, conn)
        log.debug('Deleted %s and Readded %s tags for case %s' %
                  (cd, ca, self.config.caseid))
        db.recycleConnection(conn)
        db.closeConnections()

    def add_tag(self, tagid, conn):
        sqllist = self.config.get_addsql(tagid)
        count = 0
        try:
            try:
                cur = conn.cursor()
                for sql in sqllist:
                    cur.execute(sql)
                    log.debug(cur.query)
                    for row in cur.fetchall():
                        cur.execute(ADD_TAG_SQL
                                    % {'tagid': tagid, 'caseid': row[0]})
                        log.debug(cur.query)
                        count += 1
                log.debug('Added tag "%s" for for %s cases' % (tagid, count))
                conn.commit()
            except StandardError:
                conn.rollback()
                log.exception('Error on adding tags')
        finally:
            return count

    def del_tag(self, tagid, conn=None):
        sqllist = self.config.get_delsql(tagid)
        count = 0
        try:
            try:
                cur = conn.cursor()
                for sql in sqllist:
                    cur.execute(sql)
                    log.debug(cur.query)
                    for row in cur.fetchall():
                        cur.execute(DEL_TAG_SQL
                                    % {'tagid': tagid, 'caseid': row[0]})
                        log.debug(cur.query)
                        count += 1
                log.debug('Deleted tag "%s" for for %s cases' % (tagid, count))
                conn.commit()
            except StandardError:
                conn.rollback()
                log.exception('Error on deleting tags')
        finally:
            return count
